from pathlib import Path
from argparse import Namespace

from texlite import messages as msg
from texlite.io import read_md_lines, write_tex_lines
from texlite.parse import parse
from texlite.compile import compile_tex_to_pdf


def _check_source(source_path: Path) -> (bool, str):
    '''Checks existence and file type of inputted source path'''

    # check if file exists
    if not source_path.exists():
        return False, f'Source file \'{source_path}\' not found'

    # check if file is markdown
    if not source_path.suffix == '.md':
        return False, 'Source file must be in markdown (.md) format'

    # return success and message
    return True, f'Reading in document \'{source_path}\'...'


def run(args: Namespace) -> bool:
    '''Runs main program cycle'''

    # form source path (and add .md if no suffix given)
    source_path = Path(args.source)
    if source_path.suffix == '':
        source_path = source_path.with_suffix('.md')

    # check source path
    source_success, source_message = _check_source(source_path)

    # display source success
    if source_success:
        msg.message(source_message)
    else:
        msg.error(source_message)
        return False

    # get directory path and base path (stem of files)
    dir_path = source_path.parents[0]
    base_path = dir_path / source_path.stem

    # read in markdown lines from file
    md_path = base_path.with_suffix('.md')
    md_lines = read_md_lines(path=md_path)

    # translate markdown to tex
    tex_lines = parse(md_lines,
                      graphics_path=dir_path,
                      package_config_path=args.default_packages)

    # write out tex lines to file
    tex_path = base_path.with_suffix('.tex')
    write_tex_lines(path=tex_path, lines=tex_lines)

    # compile to pdf
    if not args.no_pdf:
        compile_success, compile_message = compile_tex_to_pdf(
            path=tex_path,
            save_tex=args.save_tex,
            show_tex_output=args.show_tex_output,
        )

        # display compile success
        if compile_success:
            msg.message(compile_message)
        else:
            msg.error(compile_message)
            return False

    msg.message('Done')

    return True



from texlite.components.common import BACKSLASH


class Section:
    '''Represents a depth-aware header'''

    def __init__(self, content: str, section_type: str='section'):
        self.section_type = section_type
        self.content = content

    def tex(self) -> str:
        '''Returns generated TeX from component'''

        return f'{BACKSLASH}{self.section_type}{{{self.content}}}'




# import component classes
from .meta import Meta, DocumentBegin, DocumentEnd, MakeTitle
from .section import Section
from .text import Text
from .list import List
from .figure import Figure
from .equation import Equation



from pathlib import Path

from texlite import messages as msg
from texlite.components.common import BACKSLASH
from texlite.components.text import Text


class Figure:
    '''Represents an image, optionally with a caption'''

    def __init__(self, graphics_path: Path, image_path: Path,
                 caption_text: str='',
                 graphics_width: str=f'{BACKSLASH}textwidth'):

        # set details
        self.graphics_path = graphics_path
        self.image_path = image_path
        self.caption_text = caption_text
        self.graphics_width = graphics_width

    def tex(self) -> str:
        '''Returns generated TeX from component'''

        # check if image exists
        full_path = Path(self.graphics_path) / Path(self.image_path)
        if not Path(full_path).exists():
            msg.warning(f'Image "{full_path}" could not be found. Replacing '
                        'with placeholder text.')

            graphics = ([
                f'{BACKSLASH}texttt{{Image Not Found}}',
            ])
        else:
            graphics = [
                f'{BACKSLASH}includegraphics[width={self.graphics_width}]'
                f'{{{self.image_path}}}',
            ]

        # form figure
        if self.caption_text:
            lines = [
                f'{BACKSLASH}begin{{figure}}[h!]',
                f'{BACKSLASH}centering',
                *graphics,
                f'{BACKSLASH}caption{{{Text(self.caption_text).tex()}}}'
                f'{BACKSLASH}end{{figure}}',
            ]
        else:
            lines = [
                f'{BACKSLASH}begin{{figure}}[h!]',
                f'{BACKSLASH}centering',
                *graphics,
                f'{BACKSLASH}end{{figure}}',
            ]

        return '\n'.join(lines)



from pathlib import Path
from typing import List as L

from texlite.components.common import BACKSLASH


class Equation:
    '''Represents a full-line or multi-line mathematical equation'''

    def __init__(self, equation_lines: L[str]):

        self.equation_lines = equation_lines

    def tex(self) -> str:
        '''Returns generated TeX from component'''

        # form equation section
        lines = [
            f'{BACKSLASH}begin{{align*}}',
            *[f'{line} {BACKSLASH*2}' for line in self.equation_lines],
            f'{BACKSLASH}end{{align*}}'
        ]

        return '\n'.join(lines)



from typing import List as L

from texlite.components.common import BACKSLASH
from texlite.components.text import Text


class List:
    '''Represents recursively nestable ordered or unordered list'''

    def __init__(self, items: L[str], ordered: bool=False):

        # set details
        self.items = items
        self.ordered = ordered

    def tex(self) -> str:
        '''Returns generated TeX from component'''

        lines = []

        # add begin
        if self.ordered:
            lines.append(f'{BACKSLASH}begin{{enumerate}}')
        else:
            lines.append(f'{BACKSLASH}begin{{itemize}}')

        # add items
        for item in self.items:

            # parse nested list
            if type(item) == type(self):
                lines.append(f'{item.tex()}')

            # parse list item
            else:
                lines.append(f'{BACKSLASH}item {Text(item).tex()}')

        # add end
        if self.ordered:
            lines.append(f'{BACKSLASH}end{{enumerate}}')
        else:
            lines.append(f'{BACKSLASH}end{{itemize}}')

        return '\n'.join(lines)



import json
import os
from platform import system
from pathlib import Path
from typing import Union, List as L

from texlite import messages as msg


def get_os_name() -> str:
    '''Returns if OS is linux, windows, or macos'''

    if system() == 'Linux':
        return 'linux'
    elif system() == 'Windows':
        return 'windows'
    elif system() == 'Darwin':
        return 'macos'

    # default to linux
    return 'linux'


def read_json(path: Path) -> Union[dict, list]:
    '''Reads in JSON (.json) file as a dictionary or list'''

    try:
        with open(Path(path)) as f:
            return json.load(f)
    except FileNotFoundError:
        msg.error(f'JSON file \'{path}\' not found', halt=True)


def read_file_as_list(path: Path) -> L[str]:
    '''Reads a text (.txt) file and return list of file lines'''

    try:
        with open(Path(path)) as f:
            return [line.strip('\n') for line in f.readlines()]
    except FileNotFoundError:
        msg.error(f'File \'{path}\' not found', halt=True)



from typing import Union


# useful constants and character sets
BACKSLASH = r'\_'[0]
BANNER_LINE = r'% ' + '-' * 78
SPECIAL_CHARS = r'\#$%^&_{}~'
NON_ENCAPSULATION_CHARS = SPECIAL_CHARS + r' '
FONT_SIZES = ['8pt', '9pt', '10pt', '11pt', '12pt', '14pt', '17pt', '20pt']

# text regex patterns
BOLD_RE = r'\*\*([^\*\*]*)\*\*'
ITALICS_RE = r'\*([^\*]*)\*'
CODE_RE = r'`([^`]*)`'
QUOTES_RE = r'"([^"]*)"'
HYPERLINK_RE = r'\[([^\[]*)]\(([^\(]*)\)'


def is_number(value: Union[str, float, int]) -> bool:
    '''Returns if a value can be converted to integer or float'''

    try:
        float(value)
        return True

    except ValueError:
        return False



from pathlib import Path
from typing import Optional, List as L

from texlite.components.common import (
    is_number, BACKSLASH, BANNER_LINE, FONT_SIZES
)
from texlite import messages as msg
from texlite.utils import read_file_as_list


DEFAULT_PACKAGES = [
    'hyperref',
    'amsmath',
    'amssymb',
]


class Meta:
    '''Handles the meta options and setup of the document'''

    def __init__(self, title: Optional[str]=None,
                 author: Optional[str]=None,
                 date: Optional[str]=None,
                 abstract: Optional[str]=None,
                 fontsize: str='10pt',
                 margin: str='1.6in',
                 linespread: int=1.0,
                 usepackages: Optional[str]=None,
                 package_config_path: Optional[Path]=None,
                 graphics_path: Optional[Path]=None):

        # set default packages
        self.packages = DEFAULT_PACKAGES

        # set default packages
        if package_config_path:

            # use custom list
            self.packages += read_file_as_list(package_config_path)

        # declare specifiable meta options
        # NOTE: validation of options handled in `Meta._validate_options`
        # by default, defaults are None
        self.options = [
            'title',
            'author',
            'date',
            'abstract',
            'fontsize', # default: 10pt
            'margin', # default: 1.6in
            'linespread', # default: 1.0
            'usepackages',
        ]

        # document detail options
        self.title = title
        self.author = author
        self.date = date
        self.abstract = abstract

        # document setup options
        self.fontsize = fontsize
        self.margin = margin
        self.linespread = linespread

        # other
        self.usepackages = usepackages

        # graphics setup
        self.graphics_path = graphics_path

    def tex(self) -> str:
        '''Returns generated TeX from component'''

        # validate options
        self._validate_options()

        # add meta preface
        lines = [
            r'% meta',
            f'{BACKSLASH}documentclass[{self.fontsize}]{{extarticle}}',
        ]

        # add packages
        lines += [
            r'',
            r'% packages',
            *self._packages(),
        ]

        # add preamble commands
        lines += [
            r'',
            r'% preamble commands',
            *self._preamble_commands(),
        ]

        # add optional title details
        lines += [
            r'',
            r'% document details',
            *self._document_details()
        ]

        # return joined string
        return '\n'.join(lines)

    def _validate_options(self) -> None:
        '''Validates options to provide warnings and reset to defaults'''

        # check fontsize
        if self.fontsize not in FONT_SIZES:

            # show warning and enact default
            msg.warning(
                'Option \'fontsize\' must be one of [8pt, 9pt, 10pt, 11pt, '
                '12pt, 14pt, 17pt, 20pt]. Defaulting to 10pt.'
            )
            self.fontsize = '10pt' # reset to default

        # # check margin
        if (not is_number(self.margin[:-2]) or
                not self.margin[-2:] in ['mm', 'cm', 'pt', 'in']):

            # show warning and enact default
            msg.warning(
                'Option \'margin\' must be a number followed by one of [mm, '
                'cm, pt, in] (e.g. 0.8in). Defaulting to 1.6in.'
            )
            self.margin = '1.6in'

        # check linespread
        if not is_number(self.linespread):

            # show warning and enact default
            msg.warning(
                'Option \'linespread\' must be a float (e.g. 1.6). '
                'Defaulting to 1.0.'
            )
            self.linespread = 1.0

    def _packages(self) -> L[str]:
        '''Returns list of TeX import commands for packages'''

        lines = []

        # include encoding specification
        lines.append(f'{BACKSLASH}usepackage[utf8]{{inputenc}}')

        # include margins
        lines.append(f'{BACKSLASH}usepackage[margin={self.margin}]'
                     f'{{geometry}}')

        # include default packages
        for package in self.packages:
            lines.append(f'{BACKSLASH}usepackage{{{package}}}')

        # include extra packages
        if self.usepackages:
            extra_packages = self.usepackages.replace(' ', '').split(',')
            for package in extra_packages:
                lines.append(f'{BACKSLASH}usepackage{{{package}}} % custom')

        return lines

    def _preamble_commands(self) -> L[str]:
        '''Returns list of TeX comamands for the document preamble'''

        lines = []

        # set line spacing
        lines.append(f'{BACKSLASH}linespread{{{self.linespread}}}')

        # include path for graphics
        if self.graphics_path:

            # ensure path is POSIX
            self.graphics_path = Path(self.graphics_path).as_posix()

            lines.append(f'{BACKSLASH}usepackage{{graphicx}}')
            lines.append(f'{BACKSLASH}graphicspath'
                         f'{{{{{self.graphics_path}/}}}}')

        return lines

    def _document_details(self) -> L[str]:
        '''Returns list of TeX document detail specification commands'''

        lines = []

        # add title if applicable
        if self.title:
            lines.append(f'{BACKSLASH}title{{{BACKSLASH}'
                         f'textbf{{{self.title}}}}}')

        # add author if applicable
        if self.author:
            lines.append(f'{BACKSLASH}author{{{self.author}}}')

        # add date, defaulting an empty (unshown) date
        if self.date:
            lines.append(f'{BACKSLASH}date{{{self.date}}}'),
        else:
            lines.append(f'{BACKSLASH}date{{}}')

        return lines


class DocumentBegin:
    '''Opens the main document section'''

    def tex(self) -> str:
        '''Returns generated TeX from component'''

        return '\n'.join([
            BANNER_LINE,
            r'\begin{document}',
            BANNER_LINE,
        ])


class DocumentEnd:
    '''Closes the main document section'''

    def tex(self) -> str:
        '''Returns generated TeX from component'''

        return '\n'.join([
            BANNER_LINE,
            r'\end{document}',
            BANNER_LINE,
        ])


class MakeTitle:
    '''Creates the title from the meta document details'''

    def __init__(self, meta: Meta):
        self.title = meta.title
        self.author = meta.author
        self.date = meta.date
        self.abstract = meta.abstract

    def tex(self) -> str:
        '''Returns generated TeX from component'''

        lines = []

        if self.title:
            lines.append(f'{BACKSLASH}maketitle{{}}')

        if self.abstract:
            lines += [
                f'{BACKSLASH}begin{{abstract}}',
                f'{self.abstract}',
                f'{BACKSLASH}end{{abstract}}'
            ]

        return '\n'.join(lines)



import os
import subprocess
from pathlib import Path
from typing import List

from texlite import messages as msg
from texlite.utils import get_os_name


AUXILLARY_FILE_EXTENSIONS = ['aux', 'log', 'out']


def compile_tex_to_pdf(path: Path, save_tex: bool=False,
                       show_tex_output: bool=False) -> (bool, str):
    '''Compiles a PDF from a TeX (.tex) file'''

    msg.message("Compiling...")

    # get base file path (tex file path without extension)
    base_path = Path(path).parents[0] / Path(path).stem
    file_stem = base_path.stem

    # remove previous PDF file if exists
    if Path(f'{base_path}.pdf').exists():
        Path(f'{base_path}.pdf').unlink()

    # run pdflatex
    try:
        exit_code = _call_pdflatex(base_path, show_tex_output=show_tex_output)
    except FileNotFoundError:
        return False, ('TeX PDF compiler could not be found. If not '
                       'installed, please install a TeX distribution '
                       '(https://www.latex-project.org/get).')

    # handle pdflatex errors
    if exit_code == 1:

        _tex_clean_up(file_stem, base_path, AUXILLARY_FILE_EXTENSIONS,
                      save_tex=False)

        return False, ('TeX could not be compiled, possibly due to the '
                       'inclusion of a special character or undefined '
                       'command, or due to missing packages. Use '
                       '--show-tex-output for details.')

    # move pdf to destination
    Path(f'{file_stem}.pdf').rename(Path(f'{base_path}.pdf'))

    # clean up
    _tex_clean_up(file_stem, base_path, AUXILLARY_FILE_EXTENSIONS,
                  save_tex=save_tex)

    # NOTE: this feature is prone to errors and hard to test
    # # open PDF with program
    # if open_with:
    #     msg.message(f'Opening "{base_path}.pdf" with "{open_with}"...')
    #     exit_code = _call(f'{open_with} {base_path}.pdf')
    #     if exit_code == 127:
    #         msg.error(f'Could not open "{base_path}.pdf" with "{open_with}"')

    return True, f'Saving PDF \'{base_path}.pdf\'...'


def _call_pdflatex(base_path: Path, show_tex_output: bool=False) -> int:
    '''Calls pdflatex to run the PDF compilation'''

    # set command executable
    cmd_exe = _get_pdflatex_exe()

    # set flags
    cmd_args = ['-halt-on-error', str(base_path)]

    # run pdflatex
    cmd = [cmd_exe, *cmd_args]

    if show_tex_output:
        exit_code = subprocess.call(cmd)
    else:
        with open(os.devnull, 'w') as f:
            exit_code = subprocess.call(cmd, stdout=f)

    return exit_code


def _get_pdflatex_exe() -> str:
    '''Gets the name/path of pdflatex executable in shell'''

    # XXX: the `which` method used in this function does not work in Windows,
    # so just default to pdflatex. This is a little hacky and could be
    # improved.

    if get_os_name() == 'windows':
        return 'pdflatex'

    # attempts (in order)
    exes = [
        'pdflatex',
        '/usr/bin/pdflatex',
        '/usr/local/bin/pdflatex',
        '/Library/TeX/texbin/pdflatex', # macOS
    ]

    # open devnull pipe to write subprocess calls to
    with open(os.devnull, 'w') as f:

        # iterate through attempts
        for exe in exes:

            # attempt to call which on executable
            try:
                exit_code = subprocess.call(['which', exe], stdout=f)
                if exit_code == 0:
                    return exe
            except FileNotFoundError:
                msg.warning('Search for TeX PDF compiler failed.')
                raise FileNotFoundError

    # return pdflatex as a default
    return 'pdflatex'


def _tex_clean_up(file_stem: Path, base_path: Path,
                  auxillary_file_extensions: List[str],
                  save_tex: bool=False) -> None:
    '''Cleans up auxillary files generated in PDF compilation'''

    # clean up auxillary files
    for extension in AUXILLARY_FILE_EXTENSIONS:
        if Path(f'{file_stem}.{extension}').exists():
            Path(f'{file_stem}.{extension}').unlink()

    # remove tex if not keeping
    if not save_tex:
        if Path(f'{base_path}.tex').exists():
            Path(f'{base_path}.tex').unlink()



import re

from texlite.components.common import (
    BACKSLASH, NON_ENCAPSULATION_CHARS,
    ITALICS_RE, BOLD_RE, CODE_RE, QUOTES_RE, HYPERLINK_RE
)


class Text:
    '''Represents standard auto-formatted text'''

    def __init__(self, content: str):
        self.content = content

    def tex(self) -> str:
        '''Returns generated TeX from component'''

        # flow through formatting pipes
        formatted_content = self._format_encapsulations(self.content)
        formatted_content = self._format_replacements(formatted_content)
        formatted_content = self._format_hyperlinks(formatted_content)
        formatted_content = self._format_special_characters(formatted_content)

        return formatted_content

    def _is_encapsulatable(self, text,
                           bad_chars=NON_ENCAPSULATION_CHARS) -> bool:

        if len(text) == 0:
            return False

        # return if head and tail chars are acceptable for encapsulation
        return not text[0] in bad_chars and not text[-1] in bad_chars

    def _format_encapsulations(self, text: str) -> str:

        # format bold (**)
        for match in re.findall(BOLD_RE, text):
            if self._is_encapsulatable(match):
                text = text.replace(f'**{match}**',
                                    f'{BACKSLASH}textbf{{{match}}}')

        # format italics (*)
        for match in re.findall(ITALICS_RE, text):

            if self._is_encapsulatable(match):
                text = text.replace(f'*{match}*',
                                    f'{BACKSLASH}textit{{{match}}}')

        # format code (`)
        for match in re.findall(CODE_RE, text):
            if self._is_encapsulatable(match):
                text = text.replace(f'`{match}`',
                                    f'{BACKSLASH}texttt{{{match}}}')

        # format double quotes (")
        for match in re.findall(QUOTES_RE, text):
            if self._is_encapsulatable(text):
                text = text.replace(f'"{match}"', f'``{match}"')

        return text

    def _format_replacements(self, text: str) -> str:

        # replace horizontal bars with medskips
        if text == '---':
            return r'\medskip'

        return text

    def _format_hyperlinks(self, text: str) -> str:

        # format hyperlinks
        for m in re.finditer(HYPERLINK_RE, text):

            # get href info
            link_text = m.group(1)
            link_url = m.group(2)

            # form href and replace in text
            href = f'{BACKSLASH}href{{{link_url}}}{{{Text(link_text).tex()}}}'
            text = text.replace(f'[{link_text}]({link_url})', href)

        return text

    def _format_special_characters(self, text: str) -> str:

        # replace special characters with safe (backslashed) ones
        text = text.replace('&', r'\&')

        return text



import sys
import argparse

from texlite.cli import run
from texlite._version import description, __version__


def main() -> None:
    '''Entry-point for command-line calling (texlite.__main__:main)'''

    # parse command-line arguments
    parser = argparse.ArgumentParser(
        usage='texlite [options] <source>',
        description=description,
        epilog='For more information go to https://github.com/lucrae/texlite.'
    )
    parser.add_argument('source', type=str,
                        help='path to source markdown (.md) file to parse and '
                        'compile')
    parser.add_argument('-V', '--version', action='version',
                        version=f'TeXLite {__version__}',
                        help='display version and exit')
    parser.add_argument('-t', '--save-tex', action='store_true', default=False,
                        help='save TeX (.tex) file used in compilation')
    parser.add_argument('--show-tex-output', action='store_true',
                        default=False, help='show output for TeX and pdfLaTeX '
                        'processes')
    parser.add_argument('--no-pdf', action='store_true', default=False,
                        help='save TeX (.tex) and do not generate PDF')
    parser.add_argument('--default-packages', metavar='F', help='use text '
                        '(.txt) file to specify a custom set of default '
                        'packages to use (one line per package name)')
    args = parser.parse_args()

    # run texlite
    run(args=args)


if __name__ == '__main__':

    # run command-line entry-point
    main()



import re
import os


from texlite.utils import get_os_name


class ANSI:
    '''ANSI colour interrupt codes'''

    # default/reset code
    DEFAULT = '\033[0m'

    # colour codes
    OK = '\033[92m'
    WARNING = '\033[93m'
    ERROR = '\033[91m'

    # effects
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    # GREEN = '\033[92m'
    # YELLOW = '\033[93m'
    # RED = '\033[91m'
    # MAGENTA = '\033[95m',
    # BLUE ='\033[94m',

    @classmethod
    def disable(self):
        self.DEFAULT = ''
        self.OK = ''
        self.WARNING = ''
        self.ERROR = ''
        self.BOLD = ''
        self.UNDERLINE = ''


# disable terminal styling if not on Linux
if get_os_name() != 'linux':
    ANSI.disable()


def _print(message: str, preface: str=None, col=ANSI.DEFAULT) -> None:

    # print text
    if preface:
        print(f'{col}{ANSI.BOLD}{preface}:{ANSI.DEFAULT} {message}')
    else:
        print(f'{col}{message}{ANSI.DEFAULT}')


def message(message: str) -> None:
    '''Prints a standard message'''

    # print message
    _print(message, preface='TeXLite', col=ANSI.OK)


def warning(message: str) -> None:
    '''Prints a warning message'''

    # print warning message
    _print(message, preface='Warning', col=ANSI.WARNING)


def error(message: str, halt: bool=False) -> None:
    '''Prints an error message'''

    # print error message
    _print(message, preface='Error', col=ANSI.ERROR)

    if halt:
        # halt program
        exit()



from pathlib import Path


def read_md_lines(path: Path) -> list:
    '''Returns a stripped list from the lines of a given Markdown (.md) file'''

    # create list of markdown file lines
    lines = []
    with open(path, 'r', encoding='utf-8') as f:
        for i, raw_line in enumerate(f.readlines()):

            # remove newline markers
            line = raw_line.strip('\n')

            if line:

                # append (non-empty) line
                lines.append(line)

    return lines


def write_tex_lines(path: Path, lines: list) -> None:
    '''Writes out a TeX (.tex) file given a list of the file lines'''

    # write tex lines in file
    with open(path, 'w', encoding='utf-8') as f:
        for line in lines:
            f.write(f'{line}\n')



import re
from pathlib import Path
from setuptools import setup, find_packages

from texlite._version import description, __version__


# read in description
with open(Path('docs/DESCRIPTION.md'), 'r') as f:
    long_description = f.read()


# set up package details
setup(
    name='texlite',
    version=__version__,
    author='Lucien Rae Gentil',
    author_email='lucienraegentil@gmail.com',
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/lucrae/texlite',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    entry_points={
        'console_scripts': [
            'texlite = texlite.__main__:main'
        ]
    },
    python_requires='>=3.6',
)



import re
from pathlib import Path
from typing import List as L

from texlite.components import (
    Meta, DocumentBegin, DocumentEnd, MakeTitle, Section, Text,
    List, Figure, Equation
)


# set useful constants and string sets
NUMBERS = '0123456789'
LETTERS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
UNORDERED_LIST_PREFIXES = ['-', '+', '*']
ORDERED_LIST_PREFIXES = [f'{c}.' for c in NUMBERS + LETTERS]
LIST_PREFIXES = UNORDERED_LIST_PREFIXES + ORDERED_LIST_PREFIXES
TAB = r' ' * 4 # tabs are four spaces

# set section heading codes
HEADINGS = {
    '#': 'section',
    '##': 'subsection',
    '###': 'subsubsection',
    '#*': 'section*',
    '##*': 'subsection*',
    '###*': 'subsubsection*',
}

# regex patterns
FIGURE_RE = re.compile(r'^(!\[.*\]\(.*\))$')


def parse(md_lines: L[str], graphics_path: Path=None,
          package_config_path: Path=None) -> L[str]:
    '''Parses list of Markdown (.md) lines and returns TeX (.tex) lines'''

    # set up meta object and initial components
    meta = Meta(
        graphics_path=graphics_path,
        package_config_path=package_config_path
    )
    components = [DocumentBegin()]

    # iteratively tokenise lines
    i, n_lines = 0, len(md_lines)
    while i < n_lines:

        # read in line elements
        prefix = get_line_prefix(md_lines[i])
        body = get_line_without_prefix(md_lines[i])

        # handle optional meta specifications
        if prefix in [f':{option}:' for option in meta.options]:

            # set meta attribute to specified value
            option_name = prefix[1:-1]
            setattr(meta, option_name, body)

        # handle heading/section
        elif prefix in HEADINGS.keys():

            # add section (heading) component
            components.append(Section(body, section_type=HEADINGS[prefix]))

        # handle unordered list
        elif prefix in UNORDERED_LIST_PREFIXES:

            # parse list and add list component
            list_component, i = _parse_list(i, md_lines, ordered=False)
            components.append(list_component)

        # handle ordered list
        elif prefix in ORDERED_LIST_PREFIXES:

            # parse list and add list component
            list_component, i = _parse_list(i, md_lines, ordered=True)
            components.append(list_component)

        # handle figures
        elif FIGURE_RE.match(md_lines[i]):

            # get details of figure component
            image_path = re.findall(r'\((.*)\)', md_lines[i])[0]
            caption_text = re.findall(r'\[(.*)\]', md_lines[i])[0]

            # add figure component
            components.append(Figure(graphics_path, image_path, caption_text))

        # handle equation
        elif prefix == '$$$':

            # skip to next line and then iterate through until $$$ is reached
            equation_lines = []
            i += 1
            while i < n_lines:

                # break out at end
                if md_lines[i] == '$$$':
                    break

                equation_lines.append(md_lines[i])

                i += 1

            components.append(Equation(equation_lines))

        # handle comment
        elif prefix == '<!--':

            # iterate through lines until a '-->' is reached
            while i < n_lines:
                if '-->' in md_lines[i]:
                    break
                i += 1

        # interpret as text content (by default)
        else:
            components.append(Text(md_lines[i]))

        # increment to next line
        i += 1

    # append document end and add meta to start
    components.append(DocumentEnd())
    components.insert(0, meta)

    # add maketitle if needed
    if meta.title or meta.abstract:
        components.insert(2, MakeTitle(meta))

    # convert tokenised components to lines
    tex_lines = [component.tex() + '\n' for component in components]
    return tex_lines


def _parse_list(i: int, md_lines: L[str], ordered: bool=False,
                depth: int=0) -> List:
    '''Parse Markdown lines containing a list and return List component'''

    # NOTE: Nested lists must be done with four-space indentation

    # set prefixes
    prefixes = ORDERED_LIST_PREFIXES if ordered else UNORDERED_LIST_PREFIXES

    # iterate through lines
    items = []
    while i < len(md_lines):

        # get line without tabs (at specific depth)
        line_without_tab = md_lines[i][depth * len(TAB):]

        # read in line if at proper depth
        if get_line_prefix(line_without_tab) in prefixes:

            # add line to items
            items.append(get_line_without_prefix(line_without_tab))
            i += 1

        # if tabbed inwards, create nested list (NOTE: recursive)
        elif (TAB * (depth + 1) in md_lines[i] and
                get_line_prefix(md_lines[i].lstrip()) in LIST_PREFIXES):

            # determine if nested list
            if get_line_prefix(md_lines[i].lstrip()) in ORDERED_LIST_PREFIXES:
                nested_ordered = True
            else:
                nested_ordered = False

            # parse nested list and append
            list_component, i = _parse_list(i, md_lines,
                                            ordered=nested_ordered,
                                            depth=depth + 1)
            items.append(list_component)
            i += 1

        # exit out of list
        else:

            # go back to align with end-of-loop increment and break
            i -= 1
            break

    # return list object and ending line for parsing to be picked up from
    return List(items, ordered=ordered), i


def get_line_prefix(line: str) -> str:
    '''Returns first section of a line'''

    return line.split(' ')[0]


def get_line_without_prefix(line: str) -> str:
    '''Returns line without first section'''

    return ' '.join(line.split(' ')[1:])


def is_empty(text: str) -> bool:
    '''Returns true if the given string is completely empty'''

    return bytes(text, encoding='utf-8') == b''
